#include <signal.h>
#include <poll.h>
#include <unistd.h>
#include <cstdlib>
#include <iostream>
#include "HTTPTunneler.hh"

using namespace HTTPTunnel;


// Read and write between stdin and HTTP socket
void ReadWrite(int sock)
{
	bool keepRunning = true;

	while(keepRunning) {
		// Poll stdin and socket
		struct pollfd pollStruct[] = {
			{ STDIN_FILENO, POLLIN, 0 },
			{ sock, POLLIN, 0 }
		};
		int readyfds = poll(pollStruct, 2, -1);
		if(readyfds > 0) {
			// Check events
			for(int i = 0; i < 2; ++i) {
				short revents = pollStruct[i].revents;
				if((revents & POLLERR) || (revents & POLLHUP) || (revents & POLLNVAL)) {
					if(revents & POLLERR)
						std::cerr << "Got ERR on fd ";
					else if(revents & POLLHUP)
						std::cerr << "Got HUP on fd ";
					else
						std::cerr << "Got NVAL on fd ";
					std::cerr << i << " (" << pollStruct[i].fd << ")" << std::endl;

					keepRunning = false;
					break;
				}
				else if(revents & POLLIN) {
					// Got data to read
					int fd = pollStruct[i].fd;
					char buff[1024];
					ssize_t noread = read(fd, buff, sizeof(buff));

					// Check for EOF
					if(noread == 0) {
						// Stdin closed?
						if(fd == STDIN_FILENO) {
							std::cerr << "Got EOF from STDIN - shutting down" << std::endl;
						}
						else {
							std::cerr << "Got EOF from server - shutting down" << std::endl;
						}
						close(sock);
						keepRunning = false;
					}
					else {
						// Write data
						if(fd == STDIN_FILENO)
							write(sock, buff, noread);
						else
							write(STDOUT_FILENO, buff, noread);
					}
				}
			}
		}
		else if(readyfds < 0) {
			// Poll error
			std::cerr << "Error while polling" << std::endl;
			keepRunning = false;
		}
	}
}


// Main
int main(int argc, char** argv)
{
	try {
		// Configure global settings
		HTTPTunneler* tunneler = HTTPTunneler::Get();
		//GlobalConfigBlob globalBlob(VERBOSE);
		GlobalConfigBlob globalBlob(QUIET);
		tunneler->ConfigureGlobal(globalBlob);

		// Start the tunneling library
		tunneler->Start();
		std::cerr << "Tunneling library started" << std::endl;

		// Configure client tunnel socket
		int cliSock = -1;
		SockTunClientConfigBlob cliBlob(Enums::TCP, false, "localhost", 8080);
		TunnelerID cliID = tunneler->LaunchClientSocket(cliBlob, cliSock);
		std::cerr << "Client tunneler ID: " << cliID << std::endl;
		std::cerr << "Socket fd: " << cliSock << std::endl;

		// Do some simple I/O
		ReadWrite(cliSock);

		// Wait for tunneler to tidy up and become inactive
		while(tunneler->IsTunnelerActive(cliID))
			sleep(1);

		// Remove configured tunnels
		tunneler->RemoveTunneler(cliID);
		std::cerr << "Tunnels removed" << std::endl;

		// Stop the tunneling library
		tunneler->Stop();
		std::cerr << "Tunneling library stopped" << std::endl;
	}
	catch(TunnelError& e) {
		std::cerr << "Error running tunneler: " << e.what() << std::endl;
	}

	// Tidy up and exit
	try {
		HTTPTunneler::Release();
	}
	catch(TunnelError& e) {
		std::cerr << "Error releasing tunneler: " << e.what() << std::endl;
	}

	exit(0);
}


